/*	  Copyright (C) 2000,2001,2002  Sony Computer Entertainment America
       	  
       	  This file is subject to the terms and conditions of the GNU Lesser
	  General Public License Version 2.1. See the file "COPYING" in the
	  main directory of this archive for more details.                             */

#ifndef ps2gl_h
#define ps2gl_h

#include "GL/gl.h"

/********************************************
 * types
 */

typedef long		pgl64_t;
typedef unsigned long	pglU64_t;

#ifdef __cplusplus
extern "C" {
#endif // __cplusplus

   // immBufferVertexSize is the size in vertices of the buffers used to store
   // glBegin/glEnd geometry. there are currently 2 sets of buffers:
   // vertex, normal, tex coord, and color buffers.
   extern int	pglInit( int immBufferVertexSize, int immDrawBufferQwordSize );
   extern int	pglHasLibraryBeenInitted(void);
   extern void	pglFinish(void);

   extern void	pglWaitForVU1(void);
   extern void	pglWaitForVSync(void);
   extern void	pglSwapBuffers(void);

   // gs memory allocation

   extern void pglPrintGsMemAllocation(void);
   extern int pglHasGsMemBeenInitted(void);

   // gs mem slots

   typedef unsigned int pgl_slot_handle_t;

   extern pgl_slot_handle_t pglAddGsMemSlot( int startingPage, int pageLength, unsigned int pixelMode );
   extern void pglLockGsMemSlot( pgl_slot_handle_t slot_handle );
   extern void pglUnlockGsMemSlot( pgl_slot_handle_t slot_handle );
   extern void pglRemoveAllGsMemSlots();

   // gs mem areas

   typedef unsigned int pgl_area_handle_t;

   extern pgl_area_handle_t pglCreateGsMemArea( int width, int height, unsigned int pix_format );
   extern void pglDestroyGsMemArea( pgl_area_handle_t mem_area );

   extern void pglAllocGsMemArea( pgl_area_handle_t mem_area );
   extern void pglFreeGsMemArea( pgl_area_handle_t mem_area );

   extern void pglSetGsMemAreaWordAddr( pgl_area_handle_t mem_area, unsigned int addr );

   extern void pglBindGsMemAreaToSlot( pgl_area_handle_t mem_area, pgl_slot_handle_t mem_slot );
   extern void pglUnbindGsMemArea( pgl_area_handle_t mem_area );

   extern void pglLockGsMemArea( pgl_area_handle_t mem_area );
   extern void pglUnlockGsMemArea( pgl_area_handle_t mem_area );

   extern int pglGsMemAreaIsAllocated( pgl_area_handle_t mem_area );
   extern unsigned int pglGetGsMemAreaWordAddr( pgl_area_handle_t mem_area );

   // display and draw management

   extern void pglSetDisplayBuffers( int interlaced,
				     pgl_area_handle_t frame0_mem, pgl_area_handle_t frame1_mem );
   extern void pglSetDrawBuffers( int interlaced,
				  pgl_area_handle_t frame0_mem, pgl_area_handle_t frame1_mem,
				  pgl_area_handle_t depth_mem );

   // textures

   void pglTextureFromGsMemArea( pgl_area_handle_t tex_area_handle );

   void pglBindTextureToSlot( GLuint texId, pgl_slot_handle_t mem_slot );
   void pglFreeTexture( GLuint texId );

   // geometry

   void pglNormalPointer( GLint size, GLenum type, GLsizei stride, const GLvoid *ptr );
   void pglDrawIndexedArrays( GLenum primType,
			      int numIndices, const unsigned char* indices,
			      int numVertices );

   void pglBeginImmediateGeometry( void );
   void pglEndImmediateGeometry( void );
   void pglRenderImmediateGeometry( void );
   void pglFinishRenderingImmediateGeometry( int forceImmediateStop );

   void pglBeginGeometry( void );
   void pglEndGeometry( void );
   void pglRenderGeometry( void );
   void pglFinishRenderingGeometry( int forceImmediateStop );

   void pglSetRenderingFinishedCallback( void (*cb)(void) );

   // general

   void pglEnable( GLenum cap );
   void pglDisable( GLenum cap );

   void pglSetInterlacingOffset( float yPixels );

   const char* pglGetCurRendererName();

   // custom renderers

   void pglBeginRendererDefs();
   void pglRegisterRenderer( void *renderer );
   void pglEndRendererDefs();

   // custom prim types

   void pglRegisterCustomPrimType( GLenum primType,
				   pglU64_t requirements,
				   pglU64_t rendererReqMask,
				   int mergeContiguous );

   // custom state

   void pglEnableCustom( pglU64_t flag );
   void pglDisableCustom( pglU64_t flag );

#ifdef __cplusplus
}
#endif // __cplusplus

// "capabilities" (things that can be passed to pglEnable/pglDisable

#define PGL_CLIPPING		2

// for pglFinishRendering

#define PGL_FORCE_IMMEDIATE_STOP	1
#define	PGL_DONT_FORCE_IMMEDIATE_STOP	0

// for pglSetDrawBuffers / pglSetDisplayBuffers

#define	PGL_NONINTERLACED	0
#define PGL_INTERLACED		1

// custom prim types

#define PGL_DONT_MERGE_CONTIGUOUS	0
#define PGL_MERGE_CONTIGUOUS		1

// various limits

#define PGL_MAX_CUSTOM_RENDERERS	64
#define PGL_MAX_CUSTOM_PRIM_TYPES	32

#endif // ps2gl_h
